/* vim: set sw=2: -*- Mode: C; tab-width: 2; indent-tabs-mode: t; c-basic-offset: 2; c-indent-level: 2-*- */
/**
 * GMyth Library
 *
 * @file gmyth/gmyth_file_transfer.c
 * 
 * @brief <p> GMythFileTransfer deals with the file streaming media remote/local
 * transfering to the MythTV frontend.
 *
 * Copyright (C) 2006 INdT - Instituto Nokia de Tecnologia.
 * @author Rosfran Lins Borges <rosfran.borges@indt.org.br>
 *
 *//*
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * GStreamer MythTV plug-in properties:
 * - location (backend server hostname/URL) [ex.: myth://192.168.1.73:28722/1000_1092091.nuv]
 * - path (qurl - remote file to be opened)
 * - port number *   
 */

#include "gmyth_file_transfer.h"
#include "gmyth_uri.h"
#include "gmyth_livetv.h"
#include "gmyth_util.h"
#include "gmyth_socket.h"
#include "gmyth_context.h"
#include "gmyth_stringlist.h"

#include <unistd.h>
#include <glib.h>

#include <arpa/inet.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <errno.h>
#include <stdlib.h>

#define GMYTHTV_QUERY_HEADER 			"QUERY_FILETRANSFER"
#define GMYTHTV_RECORDER_HEADER 	"QUERY_RECORDER"

/* default values to the file transfer parameters */
#define GMYTHTV_USER_READ_AHEAD			TRUE
#define GMYTHTV_RETRIES							-1
#define GMYTHTV_FILE_SIZE						0

#define GMYTHTV_BUFFER_SIZE					128*1024

#define GMYTHTV_VERSION							30

#define GMYTHTV_TRANSFER_MAX_WAITS	700

#ifdef GMYTHTV_ENABLE_DEBUG
#define GMYTHTV_ENABLE_DEBUG				1
#else
#undef GMYTHTV_ENABLE_DEBUG
#endif

/* this NDEBUG is to maintain compatibility with GMyth library */
#ifndef NDEBUG
#define GMYTHTV_ENABLE_DEBUG				1
#endif

static guint wait_to_transfer = 0;

static gboolean has_io_access = TRUE;

enum myth_sock_types {
  GMYTH_PLAYBACK_TYPE = 0,
  GMYTH_MONITOR_TYPE,
  GMYTH_FILETRANSFER_TYPE,
  GMYTH_RINGBUFFER_TYPE
};

//static GStaticMutex mutex = G_STATIC_MUTEX_INIT;

static GMutex*			mutex 					 = NULL;

static GCond*				io_watcher_cond  = NULL;

static GMainContext *io_watcher_context = NULL;

static void gmyth_file_transfer_class_init          (GMythFileTransferClass *klass);
static void gmyth_file_transfer_init                (GMythFileTransfer *object);

static void gmyth_file_transfer_dispose  (GObject *object);
static void gmyth_file_transfer_finalize (GObject *object);

static GMythSocket *gmyth_connect_to_transfer_backend( GMythFileTransfer **transfer, guint sock_type );
static gboolean myth_init_io_watchers( GMythFileTransfer *transfer );

void gmyth_file_transfer_close( GMythFileTransfer *transfer );

static gboolean myth_control_acquire_context( gboolean do_wait );

static gboolean myth_control_release_context( );

G_DEFINE_TYPE(GMythFileTransfer, gmyth_file_transfer, G_TYPE_OBJECT)

#if 0
static guint64
mmyth_util_decode_long_long( GMythStringList *strlist, guint offset  )
{

  guint64 ret_value = 0LL;

  g_return_val_if_fail( strlist != NULL, ret_value );

  if ( offset < gmyth_string_list_length( strlist ))
    g_printerr( "[%s] Offset is lower than the GMythStringList (offset = %d)!\n", __FUNCTION__, offset );
  g_return_val_if_fail( offset < gmyth_string_list_length( strlist ), ret_value );

  gint l1 = gmyth_string_list_get_int( strlist, offset );
  gint l2 = gmyth_string_list_get_int( strlist, offset + 1 );

  ret_value = ((guint64)(l2) & 0xffffffffLL) | ((guint64)(l1) << 32);

  return ret_value;

}
#endif

static void
gmyth_file_transfer_class_init (GMythFileTransferClass *klass)
{
  GObjectClass *gobject_class;

  gobject_class = (GObjectClass *) klass;

  gobject_class->dispose  = gmyth_file_transfer_dispose;
  gobject_class->finalize = gmyth_file_transfer_finalize;
}

  static void
gmyth_file_transfer_init (GMythFileTransfer *gmyth_file_transfer)
{ 
  g_return_if_fail( gmyth_file_transfer != NULL );
  gmyth_file_transfer->mythtv_version = GMYTHTV_VERSION;
  
  /* it is used for signalizing the event socket consumer thread */
  io_watcher_cond = g_cond_new();
  
  /* mutex to control access to the event socket consumer thread */
  mutex = g_mutex_new();
}

static void
gmyth_file_transfer_dispose  (GObject *object)
{
  GMythFileTransfer *gmyth_file_transfer = GMYTH_FILE_TRANSFER(object);

  gmyth_file_transfer_close( gmyth_file_transfer );

  G_OBJECT_CLASS (gmyth_file_transfer_parent_class)->dispose (object);
}

static void
gmyth_file_transfer_finalize (GObject *object)
{
  g_signal_handlers_destroy (object);

  G_OBJECT_CLASS (gmyth_file_transfer_parent_class)->finalize (object);
}

  GMythFileTransfer*
gmyth_file_transfer_new (gint num, GString *uri_str, gshort port, gint mythtv_version)
{
  GMythFileTransfer *transfer = GMYTH_FILE_TRANSFER ( g_object_new (
	GMYTH_FILE_TRANSFER_TYPE, FALSE ));

  transfer->mythtv_version = MYTHTV_CANNOT_NEGOTIATE_VERSION;

  transfer->card_id = num;

  transfer->rec_id = -1;

  transfer->recordernum = num;
  transfer->uri = gmyth_uri_new ( uri_str->str );

  transfer->hostname = g_string_new( gmyth_uri_gethost(transfer->uri) );
  g_print( "\t--> transfer->hostname = %s\n", transfer->hostname->str );

  if ( port >= 0 )
    transfer->port = port;
  else
    transfer->port = gmyth_uri_getport( transfer->uri );

  g_print( "\t--> transfer->port = %d\n", transfer->port );

  transfer->readposition = 0;
  transfer->filesize = GMYTHTV_FILE_SIZE;
  transfer->timeoutisfast = FALSE;

  transfer->userreadahead = GMYTHTV_USER_READ_AHEAD;
  transfer->retries = GMYTHTV_RETRIES;  

  transfer->live_tv = FALSE;
  
  transfer->backend_msgs = g_hash_table_new( g_int_hash, g_int_equal );

  transfer->query = g_string_new( GMYTHTV_QUERY_HEADER );
  g_string_append_printf ( transfer->query, " %d", transfer->recordernum );
  g_print( "\t--> transfer->query = %s\n", transfer->query->str );

  transfer->control_sock = NULL;
  transfer->event_sock = NULL;
  transfer->sock = NULL;

  return transfer;
}

gboolean
gmyth_file_transfer_livetv_setup( GMythFileTransfer **transfer, GMythSocket *live_socket )
{
	(*transfer)->sock = live_socket;
	g_object_ref( live_socket );

	return TRUE;
}

gboolean
gmyth_file_transfer_playback_setup( GMythFileTransfer **transfer, gboolean live_tv )
{

  gboolean ret = TRUE;

  (*transfer)->live_tv = live_tv;

  printf("[%s] Running config to the %s...\n", __FUNCTION__, live_tv ? "LiveTV" : "FileTransfer" );

  /* configure the control socket */
  if ((*transfer)->control_sock == NULL) { 

    if ( gmyth_connect_to_transfer_backend ( transfer, GMYTH_PLAYBACK_TYPE ) == NULL ) {
      g_printerr( "Connection to backend failed (Control Socket).\n" );
      ret = FALSE;
    }

  } else {
    g_warning("Remote transfer control socket already created.\n");
  }

  return ret;

}

gboolean
gmyth_file_transfer_setup( GMythFileTransfer **transfer, gboolean live_tv )
{
  GMythStringList *strlist = NULL;

  gboolean ret = TRUE;

  (*transfer)->live_tv = live_tv;

  printf("[%s] Running config to the %s...\n", __FUNCTION__, live_tv ? "LiveTV" : "FileTransfer" );

  /* configure the control socket */
  if ((*transfer)->event_sock == NULL) { 

    if ( gmyth_connect_to_transfer_backend ( transfer, GMYTH_MONITOR_TYPE ) == NULL ) {
      g_printerr( "Connection to backend failed (Event Socket).\n" );
      ret = FALSE;
    }

  } else {
    g_warning("Remote transfer control socket already created.\n");
  }

  /* configure the socket */
  if ( (*transfer)->sock == NULL ) { 

    //if ( live_tv == FALSE ) {

    if ( gmyth_connect_to_transfer_backend ( transfer, GMYTH_FILETRANSFER_TYPE ) == NULL ) {
      g_printerr ("Connection to backend failed (Raw Transfer Socket).\n");
      ret = FALSE;
    }
    
    GIOChannel* io_channel = (*transfer)->sock->sd_io_ch, 
    			*io_channel_control = (*transfer)->control_sock->sd_io_ch;
    GIOCondition io_cond, io_cond_control;
    GIOStatus io_status = G_IO_STATUS_NORMAL;
    GError *error = NULL;
    gchar *trash = NULL;
    guint bytes_read = 0;    
    
    io_cond = g_io_channel_get_buffer_condition( io_channel );    
	
	  if ( ( io_cond & G_IO_IN ) != 0 ) {
	    do 
	    {
	      trash = g_new0( gchar, GMYTHTV_BUFFER_SIZE );
	
	      io_status = g_io_channel_read_chars( io_channel, trash, 
	      		GMYTHTV_BUFFER_SIZE, &bytes_read, &error);
	
	      g_print( "[%s] cleaning buffer on IO binary channel... %d bytes gone!\n", 
	      		__FUNCTION__, bytes_read );
	      
	      if ( trash != NULL )
	      	g_free( trash );
	      	
	      io_cond = g_io_channel_get_buffer_condition( io_channel );
	
	    } while ( ( io_cond & G_IO_IN ) != 0 && ( io_status != G_IO_STATUS_ERROR ) && (error == NULL) );
	
	  }
	  
	  io_cond_control = g_io_channel_get_buffer_condition( io_channel_control );
	  
	  if ( ( io_cond_control & G_IO_IN ) != 0 ) {
	  	g_print( "[%s] Finishind reading function: cleaning all data on the I/O control socket...\n", __FUNCTION__ );  
		  while ( gmyth_socket_read_stringlist( (*transfer)->control_sock, strlist ) > 0 )
		  {
		    if ( strlist != NULL && gmyth_string_list_length(strlist) > 0 ) 
		    {
					gint backend_code = gmyth_string_list_get_int( strlist, 0 ); // -1 on backend error
					g_print( "[%s]\t backend code = %d -\t"\
					 		"[%s prepared for reading]! (G_IO_IN) !!!\n\n", __FUNCTION__, 
		    			backend_code, ( ( io_cond & G_IO_IN ) != 0 ) ? "IS" : "IS NOT" );
		    }
		  }	
	  }

    if ( /*!(*transfer)->live_tv && */ (*transfer)->control_sock != NULL ) {
      strlist = gmyth_string_list_new();
      g_string_printf ( (*transfer)->query, "%s %d", GMYTHTV_QUERY_HEADER, (*transfer)->recordernum );

      gmyth_string_list_append_string( strlist, (*transfer)->query );
      gmyth_string_list_append_char_array( strlist, "IS_OPEN" );

      gmyth_socket_write_stringlist( (*transfer)->control_sock, strlist );
      gmyth_socket_read_stringlist( (*transfer)->control_sock, strlist );

      if ( strlist!=NULL && gmyth_string_list_get_int( strlist, 0 ) == 1 ) {
	g_print( "[%s] Remote Myth FileTransfer socket is open!\n", __FUNCTION__ );
      } else {
	g_print( "[%s] Remote Myth FileTransfer socket is CLOSED! See the MythTV Server Backend for configuration details...\n", __FUNCTION__ );
	ret = FALSE;
      }
    }

  } else {
    g_warning("Remote transfer (raw) socket already created.\n");
  }

  return ret;
}

static GMythSocket *
gmyth_connect_to_transfer_backend( GMythFileTransfer **transfer, guint sock_type )
{
  GMythSocket *sock = NULL;

  g_return_val_if_fail( transfer != NULL && *transfer != NULL, NULL );
  g_return_val_if_fail( (*transfer)->uri != NULL, NULL );

  //myth_control_acquire_context (TRUE);

  gchar *path_dir = gmyth_uri_getpath( (*transfer)->uri );

  gchar *stype = g_strdup( "" );

  sock = gmyth_socket_new();

  gmyth_socket_connect( &sock, (*transfer)->hostname->str, (*transfer)->port );

#ifdef GMYTHTV_ENABLE_DEBUG

  g_print( "[%s] --> Creating socket... (%s, %d)\n", __FUNCTION__, (*transfer)->hostname->str, (*transfer)->port );
#endif

  GMythStringList *strlist = NULL;

  GString *hostname = g_string_new( gmyth_uri_gethost( (*transfer)->uri ) );
  GString *base_str = g_string_new( "" );

  if ( (*transfer)->mythtv_version != MYTHTV_CANNOT_NEGOTIATE_VERSION || 
  			( (*transfer)->mythtv_version = gmyth_socket_check_protocol_version_number (sock, 
  					MYTHTV_VERSION_DEFAULT ) ) != MYTHTV_CANNOT_NEGOTIATE_VERSION ) {

    if (sock == NULL) {
      stype = (sock_type==GMYTH_PLAYBACK_TYPE) ? "control socket" : "file data socket";
      g_printerr( "FileTransfer, open_socket(%s): \n"
	  "\t\t\tCould not connect to server \"%s\" @ port %d\n", stype, 
	  (*transfer)->hostname->str, (*transfer)->port );
      g_object_unref(sock);
      //myth_control_release_context( );
      return NULL;
    }

    hostname = gmyth_socket_get_local_hostname();

    g_print( "[%s] local hostname = %s\n", __FUNCTION__, hostname->str  );

    if ( sock_type == GMYTH_PLAYBACK_TYPE )
    {
      (*transfer)->control_sock = sock;
      
     	g_string_printf( base_str, "ANN Playback %s %d", hostname->str, FALSE );

      gmyth_socket_send_command( (*transfer)->control_sock, base_str );
      GString *resp = gmyth_socket_receive_response( (*transfer)->control_sock );
      g_print( "[%s] Got Playback response from %s: %s\n", __FUNCTION__, base_str->str, resp->str );
      //myth_init_io_watchers ( (*transfer) );
    }
    else if ( sock_type == GMYTH_MONITOR_TYPE )
    {
      (*transfer)->event_sock = sock;
      g_string_printf( base_str, "ANN Monitor %s %d", hostname->str, TRUE );

      gmyth_socket_send_command( (*transfer)->event_sock, base_str );
      GString *resp = gmyth_socket_receive_response( (*transfer)->event_sock );
      g_print( "[%s] Got Monitor response from %s: %s\n", __FUNCTION__, base_str->str, resp->str );
      //g_thread_create( myth_init_io_watchers, (void*)(*transfer), FALSE, NULL );
      //g_thread_create( myth_init_io_watchers, (*transfer), FALSE, NULL );

      g_printerr( "[%s] Watch listener function to the IO control channel on thread %p.\n", 
      			__FUNCTION__, g_thread_self() );

    }
    else if ( sock_type == GMYTH_FILETRANSFER_TYPE )
    {
      (*transfer)->sock = sock;
      strlist = gmyth_string_list_new();
      
      if ( (*transfer)->mythtv_version > 26 )      
      	g_string_printf( base_str, "ANN FileTransfer %s %d %d", hostname->str,
        		(*transfer)->userreadahead, (*transfer)->retries );
      else
      	g_string_printf( base_str, "ANN FileTransfer %s", hostname->str );

      gmyth_string_list_append_string( strlist, base_str );
      gmyth_string_list_append_char_array( strlist, path_dir );

      gmyth_socket_write_stringlist( (*transfer)->sock, strlist );
      gmyth_socket_read_stringlist( (*transfer)->sock, strlist );

      /* socket number, where all the stream data comes from - got from the MythTV remote backend */
      (*transfer)->recordernum = gmyth_string_list_get_int( strlist, 1 );

      /* Myth URI stream file size - decoded using two 8-bytes sequences (64 bits/long long types) */
      (*transfer)->filesize = gmyth_util_decode_long_long( strlist, 2 );

      printf( "[%s] ***** Received: recordernum = %d, filesize = %" G_GUINT64_FORMAT "\n", __FUNCTION__,
	  (*transfer)->recordernum, (*transfer)->filesize );

      if ( (*transfer)->filesize <= 0 ) {
	g_print( "[%s] Got filesize equals to %llu is lesser than 0 [invalid stream file]\n", __FUNCTION__, (*transfer)->filesize );
	g_object_unref(sock);
	sock = NULL; 
      }
    }
    else if ( sock_type == GMYTH_RINGBUFFER_TYPE )
    {
      (*transfer)->sock = sock;
      //gmyth_file_transfer_spawntv( (*transfer), NULL );      

      strlist = gmyth_string_list_new();
      g_string_printf( base_str, "ANN RingBuffer %s %d", hostname->str, (*transfer)->card_id );

      gmyth_socket_send_command( (*transfer)->sock, base_str );
      GString *resp = gmyth_socket_receive_response( (*transfer)->sock );
      g_print( "[%s] Got RingBuffer response from %s: %s\n", __FUNCTION__, base_str->str, resp->str );

    }

  }

  printf("[%s] ANN %s sent: %s\n", (sock_type==GMYTH_PLAYBACK_TYPE) ? "Playback" : (sock_type==GMYTH_FILETRANSFER_TYPE) ? "FileTransfer" : "Monitor", __FUNCTION__, base_str->str);

  if ( strlist != NULL )
    g_object_unref( strlist );

  //myth_control_release_context( );

  return sock;
}    

void
gmyth_file_transfer_spawntv ( GMythFileTransfer *file_transfer, 
    GString *tvchain_id )
{
  GMythStringList *str_list;

  g_debug ("gmyth_file_transfer_spawntv.\n");

  str_list = gmyth_string_list_new ();

  g_string_printf( file_transfer->query, "%s %d", GMYTHTV_RECORDER_HEADER, 
      file_transfer->card_id );
  gmyth_string_list_append_string (str_list, file_transfer->query);
  gmyth_string_list_append_string (str_list, g_string_new ("SPAWN_LIVETV"));
  if (tvchain_id!=NULL) {
    gmyth_string_list_append_string (str_list, tvchain_id);
    gmyth_string_list_append_int (str_list, FALSE); // PIP = FALSE (0)
  }

  gmyth_socket_sendreceive_stringlist ( file_transfer->sock, str_list );

  //GString *str = NULL;

  //if (str_list!=NULL && (str = gmyth_string_list_get_string( str_list, 0 )) != NULL && strcasecmp( str->str, "ok" ) != 0 ) {
  //  g_print( "[%s]\t\tSpawnLiveTV is OK!\n", __FUNCTION__ );
  //}
  if (str_list!=NULL)
    g_object_unref (str_list);

}

gboolean
gmyth_file_transfer_is_recording ( GMythFileTransfer *file_transfer )
{
  gboolean ret = TRUE;
  
  GMythStringList *str_list = gmyth_string_list_new ();

  g_debug ( "[%s]\n", __FUNCTION__ );
  myth_control_acquire_context (TRUE);

  g_string_printf( file_transfer->query, "%s %d", GMYTHTV_RECORDER_HEADER, 
      file_transfer->rec_id >= 0 ? file_transfer->rec_id : file_transfer->card_id );
  gmyth_string_list_append_string (str_list, file_transfer->query);
  gmyth_string_list_append_string (str_list, g_string_new ("IS_RECORDING"));

  gmyth_socket_sendreceive_stringlist ( file_transfer->control_sock, str_list );

  if ( str_list != NULL && gmyth_string_list_length(str_list) > 0 ) 
  {
    GString *str = NULL;
    if ( ( str = gmyth_string_list_get_string( str_list, 0 ) ) != NULL && strcmp( str->str, "bad" )!= 0 ) {
      gint is_rec = gmyth_string_list_get_int( str_list, 0 );
      if ( is_rec != 0 )
	ret = TRUE;
      else
	ret = FALSE;
    }
  }  
  g_print( "[%s] %s, stream is %s being recorded!\n", __FUNCTION__, ret ? "YES" : "NO", ret ? "" : "NOT" );
  myth_control_release_context( );

  if ( str_list != NULL )
    g_object_unref (str_list);

  return ret;

}

gint64
gmyth_file_transfer_get_file_position ( GMythFileTransfer *file_transfer )
{
  gint64 pos = 0;

  GMythStringList *str_list = gmyth_string_list_new ();

  g_debug ( "[%s]\n", __FUNCTION__ );
  myth_control_acquire_context (TRUE);

  g_string_printf( file_transfer->query, "%s %d", GMYTHTV_RECORDER_HEADER, 
      file_transfer->rec_id >= 0 ? file_transfer->rec_id : file_transfer->card_id );

  gmyth_string_list_append_string (str_list, file_transfer->query);
  gmyth_string_list_append_string (str_list, g_string_new ("GET_FILE_POSITION"));

  gmyth_socket_sendreceive_stringlist ( file_transfer->control_sock, str_list );

  if ( str_list != NULL && gmyth_string_list_length(str_list) > 0 ) 
  {
    GString *str = NULL;
    if ( ( str = gmyth_string_list_get_string( str_list, 0 ) ) != NULL && strstr ( str->str, "bad" ) == NULL )
      pos = gmyth_util_decode_long_long( str_list, 0 );
  } 
  myth_control_release_context( );

#ifndef GMYTHTV_ENABLE_DEBUG

  g_print( "[%s] Got file position = %lld\n", __FUNCTION__, pos );
#endif
  if (str_list!=NULL)
    g_object_unref (str_list);

  return pos;

}

  glong
gmyth_file_transfer_get_recordernum( GMythFileTransfer *transfer )
{
  return transfer->recordernum;
}

  glong
gmyth_file_transfer_get_filesize( GMythFileTransfer *transfer )
{
  return transfer->filesize;
}

  gboolean
gmyth_file_transfer_isopen( GMythFileTransfer *transfer )
{
  return (transfer->sock != NULL && transfer->control_sock != NULL);
}

  void
gmyth_file_transfer_close( GMythFileTransfer *transfer )
{
  GMythStringList *strlist;

  if (transfer->control_sock == NULL)
    return;

  strlist = gmyth_string_list_new( );

  g_string_printf( transfer->query, "%s %d", GMYTHTV_QUERY_HEADER, 
      transfer->recordernum );
  gmyth_string_list_append_string( strlist, transfer->query );
  gmyth_string_list_append_char_array( strlist, "DONE" );


  if ( gmyth_socket_sendreceive_stringlist(transfer->control_sock, strlist) <= 0 )
  {
    g_printerr( "Remote file timeout.\n" );
  }

  if (transfer->sock)
  {
    g_object_unref( transfer->sock );
    transfer->sock = NULL;
  }

  if (transfer->control_sock)
  {
    g_object_unref( transfer->control_sock );
    transfer->control_sock = NULL;
  } 

}

  void
gmyth_file_transfer_reset_controlsock( GMythFileTransfer *transfer )
{
  if (transfer->control_sock == NULL)
  {
    g_printerr( "gmyth_file_transfer_reset_controlsock(): Called with no control socket" );
    return;
  }

  GString *str = gmyth_socket_receive_response( transfer->control_sock );

  g_string_free( str, TRUE );
}

void
gmyth_file_transfer_reset_sock( GMythFileTransfer *transfer )
{
  if ( transfer->sock == NULL )
  {
    g_printerr( "gmyth_file_transfer_reset_sock(): Called with no raw socket" );
    return;
  }

  GString *str = gmyth_socket_receive_response( transfer->sock );

  g_string_free( str, TRUE );
}

void
gmyth_file_transfer_reset( GMythFileTransfer *transfer )
{
  gmyth_file_transfer_reset_controlsock( transfer );
  gmyth_file_transfer_reset_sock( transfer );
}

gint64
gmyth_file_transfer_seek(GMythFileTransfer *transfer, guint64 pos, gint whence)
{
  if (transfer->sock == NULL)
  {
    g_printerr( "[%s] gmyth_file_transfer_seek(): Called with no socket", __FUNCTION__ );
    return 0;
  }

  if (transfer->control_sock == NULL)
    return 0;

  // if (!controlSock->isOpen() || controlSock->error())
  //   return 0;
  
  myth_control_acquire_context( TRUE );

  GMythStringList *strlist = gmyth_string_list_new();
  g_string_printf (transfer->query, "%s %d", GMYTHTV_QUERY_HEADER, transfer->recordernum);
  gmyth_string_list_append_string( strlist, transfer->query );
  gmyth_string_list_append_char_array( strlist, "SEEK" );
  gmyth_string_list_append_uint64( strlist, pos );
  
  gmyth_string_list_append_int( strlist, whence );  

  if (pos > 0 )
    gmyth_string_list_append_uint64( strlist, pos );
  else
    gmyth_string_list_append_uint64( strlist, transfer->readposition );
  
  gmyth_socket_sendreceive_stringlist( transfer->control_sock, strlist );

  gint64 retval = gmyth_string_list_get_int64(strlist, 0);
  transfer->readposition = retval;
  g_print( "[%s] got reading position pointer from the streaming = %lld\n", 
      __FUNCTION__, retval );

  myth_control_release_context( );

  return retval;
}

static gboolean 
myth_control_acquire_context( gboolean do_wait ) 
{
	
	gboolean ret = TRUE;	
	
	//g_mutex_lock( mutex );
	
  //while ( !has_io_access ) 
  //	g_cond_wait( io_watcher_cond, mutex );
  	
  //has_io_access = FALSE;
  
  //myth_control_acquire_context (FALSE);
  
  /*
  g_mutex_lock( mutex );
  
  if ( do_wait ) {
  	if ( !g_main_context_wait( io_watcher_context, io_watcher_cond, mutex ) )
  		ret = FALSE;
  } else if ( !g_main_context_acquire( io_watcher_context ) )
  	ret = FALSE;
  	*/
  
  return ret;
  
}

static gboolean 
myth_control_release_context( ) 
{
	
	gboolean ret = TRUE;

  //g_main_context_release( io_watcher_context );
  
  //g_mutex_unlock( mutex );
  
  //has_io_access = TRUE;

  //g_cond_broadcast( io_watcher_cond );
  
  //g_mutex_unlock( mutex );
  
 
  return ret;
  
}

static gboolean
myth_control_sock_listener( GIOChannel *io_channel, GIOCondition condition, gpointer data )
{

  GIOStatus io_status;
  GError *error = NULL;
  GIOCondition io_cond;
  gchar *trash = g_new0( gchar, GMYTHTV_BUFFER_SIZE*10 );
  guint recv = 0;
  
  myth_control_acquire_context (TRUE);
  
  gsize len = 0;
  if (condition & G_IO_HUP)
    g_print ("Read end of pipe died!\n");
    
  if ( ( condition & G_IO_IN ) != 0 ) {
  	io_status = g_io_channel_set_encoding( io_channel, NULL, &error );
    do 
    {
      //trash = g_new0( gchar, GMYTHTV_BUFFER_SIZE );

      io_status = g_io_channel_read_chars( io_channel, trash + recv, 
      		GMYTHTV_BUFFER_SIZE, &len, &error);

      g_print( "[%s] Received data buffer from IO binary channel... %d bytes gone!\n", 
      		__FUNCTION__, len );
      		
      recv += len;
      
      //msg = g_strconcat( msg, g_strdup(trash), NULL );
      
      //if ( trash != NULL )
      //	g_free( trash );
      	
      io_cond = g_io_channel_get_buffer_condition( io_channel );

    } while ( ( io_cond & G_IO_IN ) != 0 && ( io_status != G_IO_STATUS_ERROR ) );
  }
  //ret = g_io_channel_read_chars ( source, &msg, &len, NULL, &err);
  if ( io_status == G_IO_STATUS_ERROR )
    g_print ("[%s] Error reading: %s\n", __FUNCTION__, error != NULL ? error->message : "" );
  g_print ("\n[%s]\tEVENT: Read %d bytes: %s\n\n", __FUNCTION__, len, trash != NULL ? trash : "[no event data]" );
  
  myth_control_release_context( );
  
  return TRUE;

}

static gboolean
myth_init_io_watchers( GMythFileTransfer *transfer )
{
  io_watcher_context = g_main_context_default();
  //GMainLoop *loop = g_main_loop_new( io_watcher_context, FALSE );

  GSource *source;

  if ( transfer->event_sock->sd_io_ch != NULL )
    source = g_io_create_watch( transfer->event_sock->sd_io_ch, G_IO_IN | G_IO_HUP );
  else
  	goto cleanup;

  g_source_set_callback ( source, (GSourceFunc)myth_control_sock_listener, NULL, NULL );

  g_source_attach( source, io_watcher_context );

  if (source==NULL) {
    g_printerr( "[%s] Error adding watch listener function to the IO control channel!\n", __FUNCTION__ );
    goto cleanup;
  }
  
  g_print( "[%s]\tOK! Starting listener on the MONITOR event socket...\n", __FUNCTION__ );
  
  //g_main_loop_run( loop );

cleanup:
  if ( source != NULL )
    g_source_unref( source );

  return TRUE;
}	

static gboolean
gmyth_file_transfer_is_backend_message( GMythFileTransfer *transfer, 
			GMythStringList* strlist )
{	
	gboolean found_backend_msg = FALSE;	
	GString *back_msg = NULL;
	
	back_msg = gmyth_string_list_get_string( strlist, 0 );
	if ( back_msg != NULL && back_msg->str != NULL && 
					strstr( back_msg->str, "BACKEND" ) != NULL )
	{
		found_backend_msg = TRUE;
		if ( transfer->backend_msgs != NULL ) {
			
			g_hash_table_insert ( transfer->backend_msgs, 
						&(transfer->readposition), back_msg );  
			
		}
	}
	
	return found_backend_msg;
	
}

gint 
gmyth_file_transfer_read(GMythFileTransfer *transfer, GByteArray *data, gint size, gboolean read_unlimited)
{
  gint recv = 0;
  gsize bytes_read = 0;
  gint sent = 0;
  guint remaining = 0;  
  gboolean response = FALSE;

  GIOChannel *io_channel;
  GIOChannel *io_channel_control;

  GIOCondition io_cond;
  GIOCondition io_cond_control;
  GIOStatus io_status = G_IO_STATUS_NORMAL, io_status_control = G_IO_STATUS_NORMAL;   

  GMythStringList *strlist = NULL;
  GError *error = NULL;
  
  gchar *data_buffer = NULL;

  gchar *trash = g_strdup("");

  g_return_val_if_fail ( data != NULL, -2 );

  myth_control_acquire_context (FALSE);
  
  /* gets the size of the entire file, if the size requested is lesser than 0 */
  if ( size <= 0 )
    size = transfer->filesize;

  io_channel = transfer->sock->sd_io_ch;
  io_channel_control = transfer->control_sock->sd_io_ch;

  io_status = g_io_channel_set_encoding( io_channel, NULL, &error );
  if ( io_status == G_IO_STATUS_NORMAL )
    g_print( "[%s] Setting encoding to binary data socket).\n", __FUNCTION__ );

  io_cond = g_io_channel_get_buffer_condition( io_channel );  

  io_cond_control = g_io_channel_get_buffer_condition( io_channel );

  if ( transfer->sock == NULL || ( io_status == G_IO_STATUS_ERROR ) )
  {
    g_printerr( "gmyth_file_transfer_read(): Called with no raw socket.\n" );
    recv = -1;
    goto cleanup;
  }

  if ( transfer->control_sock == NULL || ( io_status_control == G_IO_STATUS_ERROR ) )
  {
    g_printerr( "gmyth_file_transfer_read(): Called with no control socket.\n" );
    recv = -1;
    goto cleanup;
  }

  wait_to_transfer = 0;

  g_string_printf ( transfer->query, "%s %d", GMYTHTV_QUERY_HEADER, 
      transfer->recordernum ); // transfer->recordernum
  g_print( "\t[%s] Transfer_query = %s\n", __FUNCTION__, transfer->query->str );
  
  sent = size;
  remaining = size - recv;  

  while ( recv < size  && !response )
  {

    strlist = gmyth_string_list_new();
    gmyth_string_list_append_char_array( strlist, transfer->query->str );
    gmyth_string_list_append_char_array( strlist, "REQUEST_BLOCK" );
    gmyth_string_list_append_int( strlist, remaining );
		gmyth_socket_write_stringlist( transfer->control_sock, strlist );
				
		/* see the number of bytes transferred by the MythTV backend */
    gmyth_socket_read_stringlist( transfer->control_sock, strlist );
    if ( strlist != NULL && gmyth_string_list_length( strlist ) > 0 ) 
    { 
    	sent = gmyth_string_list_get_int( strlist,  0 ); // -1 on backend error
    	g_print( "[%s] got SENT buffer message = %d\n", __FUNCTION__, sent );
    	if ( sent != 0 )
    	{
    		g_print( "[%s]\t received = %d bytes, backend says %d bytes sent, "\
    				"io_cond %s prepared for reading! (G_IO_IN) !!!\n\n", __FUNCTION__, 
  					recv, sent, ( ( io_cond & G_IO_IN ) != 0 ) ? "IS" : "IS NOT" );

	  		if ( sent == remaining )
	  		{
	  			g_print( "[%s]\t\tsent %d, which is equals to requested size = %d\n\n", 
	  						__FUNCTION__, sent, remaining );	  			
	  		}
	  		else
	  		{
	  			g_print( "[%s]\t\tsent %d, which is NOT equals to requested size = %d\n\n", 
	  						__FUNCTION__, sent, remaining );
	  			size = remaining = sent;
	  			if ( sent < 0 ) {
	  				goto cleanup;
	  			} 
	  		}
    	} else {
    			//continue;
  				goto cleanup;	      		
    	} // if
  	}

    guint count_bytes = 0;
      
    bytes_read = 0;
    
    data_buffer = g_new0( gchar, remaining );    

    io_status = g_io_channel_read_chars( io_channel, data_buffer, 
  		remaining, &bytes_read, &error );

    if ( bytes_read > 0 )
    {
		  recv += bytes_read;
		  count_bytes += bytes_read;
		  remaining -= bytes_read;
		  
		  /* append new data to the increasing byte array */
		  data = g_byte_array_append ( data, g_memdup( data_buffer, bytes_read ), bytes_read );
		  
		  g_print( "[%s] Reading buffer (bytes read = %d, remaining = %d)\n", __FUNCTION__, bytes_read, remaining );
		  if ( remaining == 0 ) {
		  	response = TRUE;
		  	break;
		  }

    } else {
	    if ( io_status == G_IO_STATUS_EOF ) {
				g_print( "[%s] got EOS!", __FUNCTION__ );
	    } else if ( io_status == G_IO_STATUS_ERROR ) {
				g_print( "[%s] gmyth_file_transfer_read(): socket error.\n", __FUNCTION__ );
	    }
    	break;
    }

    g_print( "[%s]\t io_cond %s prepared for reading! (G_IO_IN) !!!\n\n", __FUNCTION__, 
    		( ( io_cond & G_IO_IN ) != 0 ) ? "IS" : "IS NOT" );
	  
		if ( ( io_cond_control & G_IO_IN ) != 0 ) {
	  	g_print( "[%s] Finishind reading function: cleaning all data on the I/O control socket...\n", __FUNCTION__ );  
		  while ( gmyth_socket_read_stringlist( transfer->control_sock, strlist ) > 0 )
		  {
		    if ( strlist != NULL && gmyth_string_list_length(strlist) > 0 ) 
		    {
					gint backend_code = gmyth_string_list_get_int( strlist, 0 ); // -1 on backend error
					g_print( "[%s]\t backend code = %d -\t"\
					 		"[%s prepared for reading]! (G_IO_IN) !!!\n\n", __FUNCTION__, 
		    			backend_code, ( ( io_cond & G_IO_IN ) != 0 ) ? "IS" : "IS NOT" );
		    }
		  }
	  }
	  g_io_channel_flush( io_channel_control, NULL );
    
  } // while

cleanup:
  myth_control_release_context (mutex);
  
	if ( data_buffer != NULL )
		g_free( data_buffer );
  
  if ( trash != NULL )
    g_free( trash );

  if ( strlist != NULL )
    g_object_unref( strlist );

  g_print( "gmyth_file_transfer_read(): reqd=%d, rcvd=%d, rept=%d, "\
      "(rcvd and rept MUST be the same!)\n", size, 
      recv, sent );

  if ( error != NULL ) {
    g_printerr( "Cleaning-up ERROR: %s [msg = %s, code = %d]\n", __FUNCTION__, error->message, 
			error->code );
		g_error_free( error );
  }
  
  if ( !response ) {
  	recv = GMYTHTV_FILE_TRANSFER_READ_ERROR;
  } else {
  	transfer->readposition += recv;
  }  

  return recv;
}

void 
gmyth_file_transfer_settimeout( GMythFileTransfer *transfer, gboolean fast )
{

  GMythStringList *strlist = NULL;

  if ( transfer->timeoutisfast == fast )
    return;

  if ( transfer->sock == NULL )
  {
    g_printerr( "gmyth_file_transfer_settimeout(): Called with no socket" );
    return;
  }

  if ( transfer->control_sock == NULL )
    return;

  strlist = gmyth_string_list_new(); 
  gmyth_string_list_append_string( strlist, transfer->query );
  gmyth_string_list_append_char_array( strlist, "SET_TIMEOUT" );
  gmyth_string_list_append_int( strlist, fast ); 

  gmyth_socket_write_stringlist( transfer->control_sock, strlist );
  gmyth_socket_read_stringlist( transfer->control_sock, strlist );

  transfer->timeoutisfast = fast;

}

#ifdef DO_TESTING

  int
main( int argc, char *argv[] )
{
  g_type_init();

  GMythFileTransfer *file_transfer = gmyth_file_transfer_new( 1, 
      g_string_new("myth://192.168.1.109:6543/jshks.nuv"), -1, GMYTHTV_VERSION );
  gmyth_file_transfer_setup( &file_transfer );
  gchar *data = g_strdup("");

  gint num = gmyth_file_transfer_read( file_transfer, data, -1 );

  return 0;

}

#endif
